# Secure ORAM-based Tensor Storage with GPT-4o-mini

This project uses Path ORAM (Oblivious RAM) for securely storing and retrieving PyTorch tensors. It utilizes a C++ implementation of Path ORAM to ensure that tensor data remains secure while interacting with AI models via `GPT-4o-mini`. The Python script manages the ORAM operations and integrates the AI response generation. 

## Files in the Repository

- **`model.py`**: The main Python script that compiles the ORAM code, stores tensors securely, retrieves them, and generates responses using GPT-4o-mini.
- **`comparison.py`**: A secondary script that performs comparisons (optional use case).
- **`measure_leakage.py`**: A script to measure data leakage during the ORAM operations (optional use case).
- **`path_oram.cpp`**: The C++ implementation of Path ORAM that securely handles tensor storage and retrieval.
- **`Makefile`**: A simple Makefile to compile the ORAM C++ code and clean up executables and `.db` files.

## Requirements

- **Python** 3.x
- **PyTorch** for tensor handling
- **Langchain and OpenAI** for interacting with GPT-4o-mini
- **C++** with a compiler supporting C++17 for compiling the ORAM implementation

### Installation

1. **Install Python dependencies**:
   ```bash
   pip install torch langchain_openai

# Secure ORAM-based Tensor Storage with GPT-4o-mini

This project uses Path ORAM (Oblivious RAM) for securely storing and retrieving PyTorch tensors. It utilizes a C++ implementation of Path ORAM to ensure that tensor data remains secure while interacting with AI models via `GPT-4o-mini`. The Python script manages the ORAM operations and integrates the AI response generation. 

## Files in the Repository

- **`model.py`**: The main Python script that compiles the ORAM code, stores tensors securely, retrieves them, and generates responses using GPT-4o-mini.
- **`comparison.py`**: A secondary script that performs comparisons (optional use case).
- **`measure_leakage.py`**: A script to measure data leakage during the ORAM operations (optional use case).
- **`path_oram.cpp`**: The C++ implementation of Path ORAM that securely handles tensor storage and retrieval.
- **`Makefile`**: A simple Makefile to compile the ORAM C++ code and clean up executables and `.db` files.

## Requirements
Take a look at **requirements.txt** file to install necessary packages and libraries.

### Installation

1. **Install Python dependencies**:
   ```bash
   pip install torch langchain_openai
   ```
2. **Install C++ Dependencies**
You need a C++17 compatible compiler like `g++` to compile the ORAM C++ code.

## Compile the ORAM C++ Code

The ORAM C++ code is automatically compiled when you run the Python script. It requires a Makefile to clean up and rebuild the code. If you need to manually rebuild, use the following command:

```bash
make rebuild
```
Ensure g++ is installed for compiling the C++ code. If you're using a Windows machine, ensure you have MinGW or another suitable C++ compiler.

## Usage
### Running the Main Script (model.py)
To run the main functionality, execute the model.py script. This will:

    1. **Compile the ORAM C++ code (if not already compiled).
    2. Store PyTorch tensors securely using the ORAM implementation.
    3. Retrieve the stored tensors.
    4. Query GPT-4o-mini using the retrieved data.

``` bash
python model.py
```

You will be prompted to input a query, and the AI response will be generated based on the securely stored and retrieved tensor data.

### Optional Scripts
**comparison.py:** This script is used to compare different ORAM configurations or models. You can run it similarly by executing:

``` bash 
python comparison.py
```
**measure_leakage.py:** This script helps measure data leakage during the storage and retrieval process. It can be run with:

```bash
python measure_leakage.py
```

# Disclaimers
This project is for educational purposes and is not secure for production use. Use with caution.

