#include <iostream>
#include <unordered_map>
#include <vector>
#include <random>
#include <functional>
#include <algorithm>

using namespace std;

class ORAM {
private:
    unordered_map<int, string> storage; // Stores encrypted model parameters
    vector<int> accessHistory; // Stores access pattern history
    hash<string> str_hash; // Hash function for obfuscation
    
public:
    ORAM() {}

    // Store a model parameter securely with hashing
    void store(int key, string value) {
        size_t hashed_key = str_hash(to_string(key));
        storage[hashed_key] = value;
        accessHistory.push_back(hashed_key);
    }

    // Retrieve a model parameter securely, avoiding direct access pattern leaks
    string retrieve(int key) {
        size_t hashed_key = str_hash(to_string(key));
        if (storage.find(hashed_key) != storage.end()) {
            accessHistory.push_back(hashed_key);
            return storage[hashed_key];
        }
        return "Parameter not found";
    }

    // Simulate ORAM shuffle to obfuscate access patterns
    void shuffle() {
    random_device rd;
    mt19937 g(rd());
    std::shuffle(accessHistory.begin(), accessHistory.end(), g);  // Use std::shuffle
}

    // Display stored parameters (for testing only, not secure in real use)
    void debug_display() {
        cout << "Stored Parameters (hashed keys):\n";
        for (const auto &pair : storage) {
            cout << "Key: " << pair.first << " | Value: " << pair.second << endl;
        }
    }
};

// Test cases for ORAM functionality

int main() {
    ORAM oram;

    cout << "Storing parameters..." << endl;
    oram.store(1, "weight_1: 0.345");
    oram.store(2, "weight_2: 0.678");
    oram.store(3, "bias_1: -0.123");
    
    cout << "Retrieving parameters..." << endl;
    cout << "Retrieved: " << oram.retrieve(1) << endl;
    cout << "Retrieved: " << oram.retrieve(2) << endl;
    cout << "Retrieved: " << oram.retrieve(3) << endl;
    cout << "Retrieved: " << oram.retrieve(4) << " (should not exist)" << endl;
    
    cout << "Shuffling access patterns..." << endl;
    oram.shuffle();
    
    cout << "Displaying stored data (hashed keys)..." << endl;
    oram.debug_display();
    
    return 0;
}
