from langchain_openai import ChatOpenAI, OpenAIEmbeddings
import numpy as np

def get_embedding(text, api_key):
    """Generate embeddings using OpenAI's text-embedding model."""
    embedding_model = OpenAIEmbeddings(model="text-embedding-ada-002", openai_api_key=api_key)
    return embedding_model.embed_query(text)

def find_best_match(query, documents, api_key):
    """Find the document that is most similar to the query."""
    query_embedding = get_embedding(query, api_key)
    doc_embeddings = [get_embedding(doc, api_key) for doc in documents]

    # Compute cosine similarity
    similarities = [np.dot(query_embedding, doc) / (np.linalg.norm(query_embedding) * np.linalg.norm(doc)) for doc in doc_embeddings]
    
    # Find the index of the most similar document
    best_idx = np.argmax(similarities)
    return documents[best_idx]

def generate_response(context, query, api_key):
    """Pass best matching context and query to GPT-4o-mini."""
    chat_model = ChatOpenAI(model="gpt-4o-mini", openai_api_key=api_key)
    
    messages = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": f"Question: {query}\nContext: {context}\nAnswer:"}
    ]

    response = chat_model.invoke(messages)
    return response

def main():
    api_key = ""

    documents = [
        "Oblivious RAM (ORAM) prevents adversaries from learning access patterns.",
        "RAG models improve LLMs by retrieving external information before generation.",
        "Model inversion attacks can reconstruct training data from model outputs."
    ]

    query = "How does ORAM help secure models?"
    best_match = find_best_match(query, documents, api_key)
    
    response = generate_response(best_match, query, api_key)
    print("AI Response:", response.content)

if __name__ == "__main__":
    main()
